/**************************************************************

    "iMonitor"
    By Adrian S. <>

 **************************************************************/

#include <Arduino.h>
#include <TFT_mask.h>
#include <Data.h>
#include <SPI.h>
#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include "SPIFFS.h"
#include "AsyncJson.h"

AsyncWebServer server(80);

TFT_mask maskController = TFT_mask();
Data dataController = Data();

const char *dataPath = "/data.txt";
const char *lastDataPath = "/lastData.txt";
const char *settingsPath = "/settings.txt";

const char *ssid = "iMonitor";
const char *psk = "iMonitor";

void setup()
{
  Serial.begin(115200);

  WiFi.mode(WIFI_AP);
  WiFi.softAP(ssid, psk);

  // Initialize TFT-Display
  maskController.initTFT();

  // Initialisiere iMonitor Einstellungen
  dataController.initSettings(settingsPath);

  // Checkt auf vorhandene Daten und aktualisiert den Monitor
  if (dataController.isDataFileExists(lastDataPath))
  {
    // Daten der iSpindel und die Einstellungen des iMonitor an den Display senden
    maskController.updateTFT(dataController.getData(lastDataPath), dataController.getData(settingsPath), dataController.isSDExists());
    Serial.println("Monitor updated!");
  }
  else
  {
    // no data an den Monitor senden
    maskController.setNoData();
  }

  // Serial Monitor Ausgabe iMonitor Einstellungen
  // Serial.print("iMonitor settings :");
  // Serial.println(settingsController.getSettings());

  // Initialize SPIFFS
  if (!SPIFFS.begin(true))
  {
    Serial.println("An Error has occurred while mounting SPIFFS");
    return;
  }

  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request)
            { request->send(SPIFFS, "/index.html", "text/html"); });

  // -------------------------------------------------------------------------------

  server.on("/settings", HTTP_GET, [](AsyncWebServerRequest *request)
            { request->send(SPIFFS, "/settings.html", "text/html"); });

  // -------------------------------------------------------------------------------

  server.on("/dataprocessing", HTTP_GET, [](AsyncWebServerRequest *request)
            { request->send(SPIFFS, "/dataprocessing.html", "text/html"); });

  // -------------------------------------------------------------------------------

  server.on("/charts", HTTP_GET, [](AsyncWebServerRequest *request)
            { request->send(SPIFFS, "/charts.html", "text/html"); });

  // -------------------------------------------------------------------------------

  // Rendert Daten in der index.html
  server.on("/values", HTTP_GET, [](AsyncWebServerRequest *request)
            {
    String json = dataController.getDataForDashboard(settingsPath, lastDataPath); 
    
    // Serial Monitor ausgabe der Ausgabedaten der index.html
    // Serial.print("values data: ");
    // Serial.println(json);
    
    request->send(200, "application/json", json); 
    json = String(); });

  // -------------------------------------------------------------------------------

  // Löscht alle Daten aus dem Speicher
  server.on("/delete-data", HTTP_GET, [](AsyncWebServerRequest *request)
            {
    dataController.deleteAllData(dataPath); 
    dataController.deleteAllData(lastDataPath);
    
    String html = "<!DOCTYPE html><html><head><meta http-equiv=\"refresh\" content=\"5; url=/\" /></head><body>Daten gelöscht!<br>Sie werden in 5 sek weitergeleitet...</body></html>";
    
    // no data an den Monitor senden
    maskController.setNoData();

    request->send(200, "text/html", html); });

  // -------------------------------------------------------------------------------

  // Rendert Daten in der charts.html
  server.on("/view-data", HTTP_GET, [](AsyncWebServerRequest *request)
            { dataController.requestDataFromFile(request, dataPath); });

  // -------------------------------------------------------------------------------

  // Stellt Datei zum download aus der SD-Karte
  server.on("/download-data", HTTP_GET, [](AsyncWebServerRequest *request)
            {
    bool result = dataController.isDataFileExists(dataPath);
    
    // Serial Monitor ausgabe
    // Serial.print("is file exist: ");
    // Serial.println(result);

    // Prüft ob die datei vorhanden ist          
    if (!result)
    {
      Serial.println("No file found!");
      String html = "<!DOCTYPE html><html><head><meta http-equiv=\"refresh\" content=\"5; url=/\" /></head><body>Es wurden keine Daten gefunden!<br>Sie werden in 5 sek weitergeleitet...</body></html>";
      request->send(200, "text/html", html);
    } else {
    dataController.downloadDataFromFile(request, dataPath);
    Serial.println("downloading!");
    } });

  // -------------------------------------------------------------------------------

  // Verarbeitung der eingehenden Einstellungs-Werten
  server.on("/safe-config", HTTP_GET, [](AsyncWebServerRequest *request)
            {
    // Erstellt einen JSON-Buffer
    StaticJsonDocument<200> settings;

    // Fügt Einträge zum JSON-Buffer hinzu
    settings["gravity_unit"] = request->getParam("gravity-unit")->value();
    settings["temp_unit"] = request->getParam("temp-unit")->value();

    // Erstellt den JSON-String
    String settingsString;
    serializeJson(settings, settingsString);

    // Einstellungen speichern speichern
    dataController.deleteAllData(settingsPath);
    dataController.appendData(settingsString, settingsPath);

    // Serial Monitor ausgabe
    // Serial.print("settings data: ");
    //Serial.println(settingsString);
    
    // Prüft ob Daten vorhanden sind und aktualisiert den iMonitor
    if (!dataController.isDataFileExists(lastDataPath)){
      // Monitor aktualisieren
      maskController.setNoData();
    } else {
      // Monitor aktualisieren
      maskController.updateTFT(dataController.getData(lastDataPath), settingsString, dataController.isSDExists());
    }

    String html = "<!DOCTYPE html><html><head><meta http-equiv=\"refresh\" content=\"5; url=/\" /></head><body>Einstellungen wurden gespeichert!<br>Sie werden in 5 sek weitergeleitet...</body></html>";
    request->send(200, "text/html", html); });

  // -------------------------------------------------------------------------------

  // Empfangt Daten der iSpindel
  server.on(
      "/iMonitor", HTTP_POST,
      [](AsyncWebServerRequest *request) {

      },
      NULL,
      [](AsyncWebServerRequest *request, uint8_t *data, size_t len, size_t index, size_t total)
      {
        // Serial Monitor ausgabe der eingehenden Daten der iSpindel
        // Serial.print("iSpindel data: ");
        // Serial.println((char *)data);

        // Daten der iSpindel und die Einstellungen des iMonitor an den Display senden
        maskController.updateTFT((char *)data, dataController.getData(settingsPath), dataController.isSDExists());

        // Daten der iSpindel speichern im lastDataPath
        dataController.deleteAllData(lastDataPath);
        dataController.appendData((char *)data, lastDataPath);

        // Generiere einen reduzierten String der Daten für den dataPath
        String viewData = dataController.setViewData((char *)data) + ",";

        // Reduzierte Daten hinzufügen im dataPath
        dataController.appendData(viewData, dataPath);

        // Serial Monitor Ausgabe
        Serial.print("Appending data to view:");
        Serial.println(viewData);
      });

  server.serveStatic("/", SPIFFS, "/");

  server.begin();
}

void loop()
{
}
